#!/usr/bin/env python2
# encoding: utf-8
import sys
from pyBusPirateLite.I2C import *
from pyBusPirateLite.I2Chigh import *

# based on http://ww1.microchip.com/downloads/en/DeviceDoc/22103a.pdf,
# gsupload, and nefariousdogooder's "GS transfer methods.txt"

# MCP23018 I2C device address
mcp_device_id = 0x20
mcp_read_byte = (mcp_device_id << 1) | 1
mcp_write_byte = (mcp_device_id << 1) | 0

# MCP23018 registers
IODIRA    = 0x00  # I/O direction: 0 write, 1 read
IODIRB    = 0x01
IOPOLA    = 0x02  # I/O polarity: 0 normal, 1 invert
IOPOLB    = 0x03
GPINTENA  = 0x04
GPINTENB  = 0x05
DEFVALA   = 0x06
DEFVALB   = 0x07
INTCONA   = 0x08
INTCONB   = 0x09
IOCON     = 0x0A # also 0x0B
GPPUA     = 0x0C # pullup: 0 off, 1 on
GPPUB     = 0x0D
INTFA     = 0x0E
INTFB     = 0x0F
INTCAPA   = 0x10
INTCAPB   = 0x11
GPIOA     = 0x12 # direct GPIO access
GPIOB     = 0x13
OLATA     = 0x14 # output latch
OLATB     = 0x15

def mcp_read(address):
  return i2c.get_byte(mcp_device_id, address)

def mcp_write(address, value):
  i2c.set_byte(mcp_device_id, address, value)

DEBUG = False

# my setup

def do_setup():
  global i2c
  i2c = I2Chigh("/dev/bus_pirate", 115200)
  # Bus Pirate setup
  print "Entering binmode: ",
  if i2c.BBmode():
    print "OK."
  else:
    print "failed."
    sys.exit()

  print "Entering raw I2C mode: ",
  if i2c.enter_I2C():
    print "OK."
  else:
    print "failed."
    sys.exit()
    
  print "Configuring I2C."
  if not i2c.cfg_pins(I2CPins.POWER | I2CPins.PULLUPS):
    print "Failed to set I2C peripherals."
    sys.exit()
  if not i2c.set_speed(I2CSpeed._400KHZ):
    print "Failed to set I2C Speed."
    sys.exit()
  #i.timeout(0.2)

  # MCP23018 setup

  # disable sequential operation
  # bit 5 = 1: sequential operation disabled
  # bit 7 = 0: bank = 0
  mcp_write(IOCON, 0x20)

  # A is output
  mcp_write(IODIRA, 0)
  # B is input
  mcp_write(IODIRB, 0xff)

  # pullups
  mcp_write(GPPUA, 0xff)
  mcp_write(GPPUB, 0)

def do_write(x):
  mcp_write(OLATA, x)
  if DEBUG: print "-> " + hex_string(mcp_read(OLATA))

def do_read():
  x = mcp_read(GPIOB)
  if DEBUG: print "<- " + hex_string(x)
  return x


# Gameshark stuff
def InitComms():
  print "Sync..."
  while True:
    # send two nibbles: 3 and 3, effectively '!'
    do_write(0x13)
    i2c.timeout(0.1)
    retval = do_read()>>4
    do_write(0);
    i2c.timeout(0.1)

    do_write(0x13)
    i2c.timeout(0.1)
    retval = (retval << 4) | (do_read()>>4)
    do_write(0)

    if retval == ord('g'): break

    print "retry"

  print "Handshake..."
  while True:
    if ReadWriteChar('G') != 'g': continue
    if ReadWriteChar('T') == 't': break
  print "OK"


def ReadWriteNibble(x, timeout=-1):
  do_write(x&0xf|0x10)
  spincount = 0

  while do_read() & 8 == 0:
    spincount += 1
    if timeout != -1:
      timeout -= 1
      if timeout == 0:
        print "timeout1"
        return 0

  retval = (do_read()&0xf0)>>4

  do_write(0)
  
  spincount = 0

  while do_read() & 8 == 8:
    spincount += 1
    if timeout != -1:
      timeout -= 1
      if timeout == 0:
        print "timeout2"
        return 0

  return retval

def ReadWriteByte(b, timeout=-1):
  return (ReadWriteNibble(b>>4, timeout)<<4)|ReadWriteNibble(b, timeout)

def ReadByte(timeout=-1):
  return ReadWriteByte(0, timeout)

def ReadWriteChar(c, timeout=-1):
  return chr(ReadWriteByte(ord(c), timeout))

def ReadChar(timeout=-1):
  return ReadWriteChar('\0', timeout)

def ReadWrite32(x, timeout=-1):
  return (ReadWriteNibble(x>>28,timeout)<<28)|\
         (ReadWriteNibble(x>>24,timeout)<<24)|\
         (ReadWriteNibble(x>>20,timeout)<<20)|\
         (ReadWriteNibble(x>>16,timeout)<<16)|\
         (ReadWriteNibble(x>>12,timeout)<<12)|\
         (ReadWriteNibble(x>> 8,timeout)<< 8)|\
         (ReadWriteNibble(x>> 4,timeout)<< 4)|\
         ReadWriteNibble(x,timeout)

def GSVersion():
  sig = [ReadWriteChar('f'), ReadChar(), ReadChar(), ReadChar()]

  strlen = ReadByte()
  version='' #hex_string(sig) + ': '
  for _ in range(strlen):
    version += ReadChar()

  return version


def Read(address, size):
  ReadWriteByte(1)
  ReadWrite32(address)
  ReadWrite32(size)

  b = ''
  for _ in range(size):
    b += ReadChar()
    print '.',
    sys.stdout.flush()

  print

  ReadWrite32(0)
  ReadWrite32(0)
  ReadWriteByte(0)

  return b

def Write(address, data):
  size = len(data)
  ReadWriteByte(2)
  ReadWrite32(address)
  ReadWrite32(size)

  for c in data:
    ReadWriteChar(c)
    print '.',
    sys.stdout.flush()

  print

  ReadWrite32(0)
  ReadWrite32(0)
  ReadWriteByte(0)

def Disconnect():
  InitComms()
  ReadWriteChar('d', 100)


# general debugging

def hex_string(s):
  if type(s) == int:
    return '%02x'%s
  else:
    return ''.join('%02x'%ord(c) for c in s)

def d(s):
  print(hex_string(s))


# 

if __name__ == '__main__':
  do_setup()
  
  # get the GS's attention
  InitComms()
  #print GSVersion()

  print "Reading"
  r = Read(0x80000000, 0x100)
  print hex_string(r)

  print "Writing"
  Write(0x80000000, r)
  Disconnect()

  mcp_write(GPIOA, 0)
  
  print "Reset Bus Pirate to user terminal: "
  if i2c.resetBP():
    print "OK."
  else:
    print "failed."
    sys.exit()
    
